@@ -36,6 +36,7 @@ import java.net.ServerSocket;
 import java.net.Socket;
 import java.net.URI;
 import java.net.URL;
+import java.security.AccessControlException;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collections;
@@ -146,7 +147,7 @@ public class SweetHome3D extends HomeApplication {
   private boolean                 pluginManagerInitialized;
   private boolean                 checkUpdatesNeeded;
   private AutoRecoveryManager     autoRecoveryManager;
-  private final Map<Home, JFrame> homeFrames;
+  private final Map<Home, HomeFrameController> homeFrameControllers;
 
   /**
    * Creates a home application instance. Recorders, user preferences, content
@@ -154,7 +155,7 @@ public class SweetHome3D extends HomeApplication {
    * lazily instantiated to let subclasses override their creation.
    */
   protected SweetHome3D() {
-    this.homeFrames = new HashMap<Home, JFrame>();
+    this.homeFrameControllers = new HashMap<Home, HomeFrameController>();
   }
 
   /**
@@ -320,10 +321,17 @@ public class SweetHome3D extends HomeApplication {
   }
 
   /**
-   * Returns the frame that displays a given <code>home</code>.
+   * Returns the frame that displays the given <code>home</code>.
    */
   JFrame getHomeFrame(Home home) {
-    return this.homeFrames.get(home);
+    return (JFrame)SwingUtilities.getRoot((JComponent)this.homeFrameControllers.get(home).getView());
+  }
+
+  /**
+   * Returns the controller of the given <code>home</code>.
+   */
+  HomeFrameController getHomeFrameController(Home home) {
+    return this.homeFrameControllers.get(home);
   }
 
   /**
@@ -390,7 +398,7 @@ public class SweetHome3D extends HomeApplication {
 
     // Add a listener that opens a frame when a home is added to application
     addHomesListener(new CollectionListener<Home>() {
-      private boolean        firstApplicationHomeAdded;
+      private boolean firstApplicationHomeAdded;
 
       public void collectionChanged(CollectionEvent<Home> ev) {
         switch (ev.getType()) {
@@ -404,8 +412,7 @@ public class SweetHome3D extends HomeApplication {
                 addNewHomeCloseListener(home, controller.getHomeController());
               }
 
-              JFrame homeFrame = (JFrame)SwingUtilities.getRoot((JComponent) controller.getView());
-              homeFrames.put(home, homeFrame);
+              homeFrameControllers.put(home, controller);
             } catch (IllegalStateException ex) {
               // Check exception by class name to avoid a mandatory bind to Java 3D
               if ("javax.media.j3d.IllegalRenderingStateException".equals(ex.getClass().getName())) {
@@ -418,7 +425,7 @@ public class SweetHome3D extends HomeApplication {
             }
             break;
           case DELETE:
-            homeFrames.remove(ev.getItem());
+            homeFrameControllers.remove(ev.getItem());
 
             // If application has no more home
             if (getHomes().isEmpty() && !OperatingSystem.isMacOSX()) {
@@ -440,9 +447,16 @@ public class SweetHome3D extends HomeApplication {
 
     addComponent3DRenderingErrorObserver();
 
+    getUserPreferences();
+    try {
+      // Set User Agent to follow statistics on used operating systems 
+      System.setProperty("http.agent", getId() + "/" + getVersion()  
+           + " (" + System.getProperty("os.name") + " " + System.getProperty("os.version") + "; " + System.getProperty("os.arch") + "; " + Locale.getDefault() + ")");
+    } catch (AccessControlException ex) {
+      // Ignore User Agent change
+    }
     // Init look and feel afterwards to ensure that Swing takes into account
     // default locale change
-    getUserPreferences();
     initLookAndFeel();
     try {
       this.autoRecoveryManager = new AutoRecoveryManager(this);
@@ -485,13 +499,13 @@ public class SweetHome3D extends HomeApplication {
       System.setProperty("apple.awt.application.name", applicationName);
       // Use Mac OS X screen menu bar for frames menu bar
       System.setProperty("apple.laf.useScreenMenuBar", "true");
-      // Force the use of Quartz under Mac OS X for better Java 2D rendering
-      // performance
+      // Force the use of Quartz under Mac OS X for better Java 2D rendering performance
       System.setProperty("apple.awt.graphics.UseQuartz", "true");
     }
-    // Set User Agent to follow statistics on used operating systems 
-    System.setProperty("http.agent", getId() + "/" + getVersion()  
-         + " (" + System.getProperty("os.name") + " " + System.getProperty("os.version") + "; " + System.getProperty("os.arch") + "; " + Locale.getDefault() + ")");
+    // Request to use system proxies to access to the Internet
+    if (System.getProperty("java.net.useSystemProxies") == null) {
+      System.setProperty("java.net.useSystemProxies", "true");
+    }
   }
 
   /**
@@ -499,8 +513,8 @@ public class SweetHome3D extends HomeApplication {
    */
   private void initLookAndFeel() {
     try {
-      // Apply current system look and feel
-      UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
+      // Apply current system look and feel if swing.defaultlaf isn't defined
+      UIManager.setLookAndFeel(System.getProperty("swing.defaultlaf", UIManager.getSystemLookAndFeelClassName()));
       // Change default titled borders under Mac OS X 10.5
       if (OperatingSystem.isMacOSXLeopardOrSuperior()) {
         UIManager.put("TitledBorder.border", UIManager.getBorder("TitledBorder.aquaVariant"));
