@@ -19,16 +19,21 @@
  */
 package com.eteks.sweethome3d.j3d;
 
+import java.awt.EventQueue;
 import java.awt.Shape;
 import java.awt.geom.Area;
+import java.awt.geom.GeneralPath;
 import java.awt.geom.Line2D;
 import java.awt.geom.PathIterator;
 import java.awt.geom.Point2D;
+import java.awt.geom.Rectangle2D;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collections;
 import java.util.Comparator;
 import java.util.List;
+import java.util.Map;
+import java.util.WeakHashMap;
 
 import javax.media.j3d.Appearance;
 import javax.media.j3d.BranchGroup;
@@ -39,10 +44,15 @@ import javax.media.j3d.RenderingAttributes;
 import javax.media.j3d.Shape3D;
 import javax.media.j3d.Texture;
 import javax.media.j3d.TextureAttributes;
+import javax.media.j3d.Transform3D;
+import javax.media.j3d.TransformGroup;
 import javax.media.j3d.TransparencyAttributes;
 import javax.vecmath.Point3f;
 import javax.vecmath.TexCoord2f;
+import javax.vecmath.Vector3f;
 
+import com.eteks.sweethome3d.model.Content;
+import com.eteks.sweethome3d.model.DoorOrWindow;
 import com.eteks.sweethome3d.model.Home;
 import com.eteks.sweethome3d.model.HomeEnvironment;
 import com.eteks.sweethome3d.model.HomeFurnitureGroup;
@@ -59,14 +69,18 @@ import com.sun.j3d.utils.geometry.NormalGenerator;
 public class Wall3D extends Object3DBranch {
   private static final TextureAttributes MODULATE_TEXTURE_ATTRIBUTES = new TextureAttributes();
   private static final float LEVEL_ELEVATION_SHIFT = 0.1f;
+  private static final Area  FULL_FACE_CUT_OUT_AREA = new Area(new Rectangle2D.Float(-0.5f, 0.5f, 1, 1));
   
   static {
     MODULATE_TEXTURE_ATTRIBUTES.setTextureMode(TextureAttributes.MODULATE);
   }
-  
+
   private static final int WALL_LEFT_SIDE  = 0;
   private static final int WALL_RIGHT_SIDE = 1;
   
+  private static Map<HomePieceOfFurniture, ModelRotationTuple> doorOrWindowRotatedModels = new WeakHashMap<HomePieceOfFurniture, ModelRotationTuple>();
+  private static Map<ModelRotationTuple, Area>                 rotatedModelsFrontAreas   = new WeakHashMap<ModelRotationTuple, Area>();
+  
   private final Home home;
 
   /**
@@ -80,7 +94,7 @@ public class Wall3D extends Object3DBranch {
    * Creates the 3D wall matching the given home <code>wall</code>.
    */
   public Wall3D(Wall wall, Home home, boolean ignoreDrawingMode, 
-                boolean waitTextureLoadingEnd) {
+                boolean waitModelAndTextureLoadingEnd) {
     setUserData(wall);
     this.home = home;
 
@@ -102,8 +116,8 @@ public class Wall3D extends Object3DBranch {
     }
         
     // Set wall shape geometry and appearance
-    updateWallGeometry();
-    updateWallAppearance(waitTextureLoadingEnd);
+    updateWallGeometry(waitModelAndTextureLoadingEnd);
+    updateWallAppearance(waitModelAndTextureLoadingEnd);
   }
 
   /**
@@ -146,22 +160,23 @@ public class Wall3D extends Object3DBranch {
 
   @Override
   public void update() {
-    updateWallGeometry();
+    updateWallGeometry(false);
     updateWallAppearance(false);
   }
   
   /**
    * Sets the 3D geometry of this wall shapes that matches its 2D geometry.  
    */
-  private void updateWallGeometry() {    
-    updateWallSideGeometry(WALL_LEFT_SIDE, ((Wall)getUserData()).getLeftSideTexture());
-    updateWallSideGeometry(WALL_RIGHT_SIDE, ((Wall)getUserData()).getRightSideTexture());
+  private void updateWallGeometry(boolean waitDoorOrWindowModelsLoadingEnd) {    
+    updateWallSideGeometry(WALL_LEFT_SIDE, ((Wall)getUserData()).getLeftSideTexture(), waitDoorOrWindowModelsLoadingEnd);
+    updateWallSideGeometry(WALL_RIGHT_SIDE, ((Wall)getUserData()).getRightSideTexture(), waitDoorOrWindowModelsLoadingEnd);
   }
   
-  private void updateWallSideGeometry(int wallSide, HomeTexture texture) {
-    Group [] wallSideGroups = {(Group)getChild(wallSide * 3),      // Bottom group
-                               (Group)getChild(wallSide * 3 + 1),  // Main group
-                               (Group)getChild(wallSide * 3 + 2)}; // Top group
+  private void updateWallSideGeometry(int wallSide, HomeTexture texture, 
+                                      boolean waitDoorOrWindowModelsLoadingEnd) {
+    Group [] wallSideGroups = {(Group)getChild(wallSide * 3),      // Bottom group (0 or 3)
+                               (Group)getChild(wallSide * 3 + 1),  // Main group   (1 or 4)
+                               (Group)getChild(wallSide * 3 + 2)}; // Top group    (2 or 5)
     Shape3D [] wallFilledShapes = new Shape3D [wallSideGroups.length];
     Shape3D [] wallOutlineShapes = new Shape3D [wallSideGroups.length];
     int [] currentGeometriesCounts = new int [3];
@@ -177,7 +192,8 @@ public class Wall3D extends Object3DBranch {
       List [] wallGeometries = {new ArrayList<Geometry>(), 
                                 new ArrayList<Geometry>(), 
                                 new ArrayList<Geometry>()};
-      createWallGeometries(wallGeometries [0], wallGeometries [1], wallGeometries [2], wallSide, texture);
+      createWallGeometries(wallGeometries [0], wallGeometries [1], wallGeometries [2], wallSide, 
+          texture, waitDoorOrWindowModelsLoadingEnd);
       for (int i = 0; i < wallSideGroups.length; i++) {
         for (Geometry wallGeometry : (List<Geometry>)wallGeometries [i]) {
           if (wallGeometry != null) {
@@ -204,30 +220,31 @@ public class Wall3D extends Object3DBranch {
    * that intersect wall.
    */
   private void createWallGeometries(List<Geometry> wallBottomGeometries, 
-                                    List<Geometry> wallGeometries, 
-                                    List<Geometry> wallTopGeometries, 
-                                    int wallSide, HomeTexture texture) {
-    float [][] wallSidePoints = getWallSidePoints(wallSide);
-    float [] textureReferencePoint = wallSide == WALL_LEFT_SIDE
+                                    final List<Geometry> wallGeometries, 
+                                    final List<Geometry> wallTopGeometries, 
+                                    final int wallSide, final HomeTexture texture,
+                                    final boolean waitDoorOrWindowModelsLoadingEnd) {
+    final float [][] wallSidePoints = getWallSidePoints(wallSide);
+    final float [] textureReferencePoint = wallSide == WALL_LEFT_SIDE
         ? wallSidePoints [0]
         : wallSidePoints [wallSidePoints.length - 1];
     Shape wallShape = getShape(wallSidePoints);
     Area wallArea = new Area(wallShape);
-    float wallElevation = getWallElevation();
+    final float wallElevation = getWallElevation();
     float wallHeightAtStart = getWallHeightAtStart();
     float wallHeightAtEnd = getWallHeightAtEnd();
     float maxWallHeight = Math.max(wallHeightAtStart, wallHeightAtEnd);
 
     // Compute wall angles and top line factors
-    Wall wall = (Wall)getUserData();
+    final Wall wall = (Wall)getUserData();
     double wallYawAngle = Math.atan2(wall.getYEnd() - wall.getYStart(), wall.getXEnd() - wall.getXStart()); 
-    double cosWallYawAngle = Math.cos(wallYawAngle);
-    double sinWallYawAngle = Math.sin(wallYawAngle);
+    final double cosWallYawAngle = Math.cos(wallYawAngle);
+    final double sinWallYawAngle = Math.sin(wallYawAngle);
     double wallXStartWithZeroYaw = cosWallYawAngle * wall.getXStart() + sinWallYawAngle * wall.getYStart();
     double wallXEndWithZeroYaw = cosWallYawAngle * wall.getXEnd() + sinWallYawAngle * wall.getYEnd();
     boolean roundWall = wall.getArcExtent() != null && wall.getArcExtent() != 0; 
-    double topLineAlpha;
-    double topLineBeta;
+    final double topLineAlpha;
+    final double topLineBeta;
     if (wallHeightAtStart == wallHeightAtEnd) {
       topLineAlpha = 0;
       topLineBeta = wallHeightAtStart;
@@ -238,6 +255,7 @@ public class Wall3D extends Object3DBranch {
     
     // Search which doors or windows intersect with this wall side
     List<DoorOrWindowArea> windowIntersections = new ArrayList<DoorOrWindowArea>();
+    List<HomePieceOfFurniture> intersectingDoorOrWindows = new ArrayList<HomePieceOfFurniture>();
     for (HomePieceOfFurniture piece : getVisibleDoorsAndWindows(this.home.getFurniture())) {
       float pieceElevation = piece.getGroundElevation();
       if (pieceElevation + piece.getHeight() > wallElevation
@@ -248,6 +266,7 @@ public class Wall3D extends Object3DBranch {
         intersectionArea.intersect(pieceArea);
         if (!intersectionArea.isEmpty()) {
           windowIntersections.add(new DoorOrWindowArea(intersectionArea, Arrays.asList(new HomePieceOfFurniture [] {piece})));
+          intersectingDoorOrWindows.add(piece);
           // Remove from wall area the piece shape
           wallArea.subtract(pieceArea);
         }
@@ -396,14 +415,24 @@ public class Wall3D extends Object3DBranch {
                 }
               }
               float doorOrWindowTop = highestDoorOrWindowElevation + highestDoorOrWindow.getHeight();
-              // Compute the minimum vertical position of wallPartPoints
-              double minTopY = maxWallHeight;
+              boolean generateGeometry = true;
+              // Translate points of wall part under doorOrWindowTop along sloping wall top  
               for (int i = 0; i < wallPartPoints.length; i++) {
                 double xTopPointWithZeroYaw = cosWallYawAngle * wallPartPoints[i][0] + sinWallYawAngle * wallPartPoints[i][1];
-                minTopY = Math.min(minTopY, topLineAlpha * xTopPointWithZeroYaw + topLineBeta);
-              }            
+                double topPointWithZeroYawElevation = topLineAlpha * xTopPointWithZeroYaw + topLineBeta;
+                if (doorOrWindowTop > topPointWithZeroYawElevation) {
+                  if (topLineAlpha == 0 || roundWall) {
+                    // Ignore geometry above window for flat wall or round sloping wall
+                    generateGeometry = false;
+                    break;
+                  }
+                  double translation = (doorOrWindowTop - topPointWithZeroYawElevation) / topLineAlpha;
+                  wallPartPoints [i][0] += (float)(translation * cosWallYawAngle);
+                  wallPartPoints [i][1] += (float)(translation * sinWallYawAngle);
+                }
+              }
               // Generate geometry for wall part above window
-              if (doorOrWindowTop < minTopY) {
+              if (generateGeometry) {
                 wallGeometries.add(createWallVerticalPartGeometry(wall, wallPartPoints, doorOrWindowTop, 
                     cosWallYawAngle, sinWallYawAngle, topLineAlpha, topLineBeta, texture, textureReferencePoint, wallSide));
                 wallGeometries.add(createWallHorizontalPartGeometry(
@@ -423,6 +452,94 @@ public class Wall3D extends Object3DBranch {
         }
       } 
     }
+
+    // Generate geometry around non rectangular doors and windows placed in straight walls along the same axis 
+    if (!roundWall && intersectingDoorOrWindows.size() > 0) {
+      final double epsilon = Math.PI / 720; // Quarter of a degree
+      final ArrayList<HomePieceOfFurniture> missingModels = new ArrayList<HomePieceOfFurniture>(intersectingDoorOrWindows.size());
+      // Compute geometry for doors or windows that have a known front area
+      for (final HomePieceOfFurniture doorOrWindow : intersectingDoorOrWindows) {
+        if (doorOrWindow instanceof DoorOrWindow
+            && !"M0,0 v1 h1 v-1 z".equals(((DoorOrWindow)doorOrWindow).getCutOutShape())) {
+          double angleDifference = Math.abs(wallYawAngle - doorOrWindow.getAngle()) % (2 * Math.PI);
+          if (angleDifference < epsilon
+              || angleDifference > 2 * Math.PI - epsilon
+              || Math.abs(angleDifference - Math.PI) < epsilon) {
+            final int frontOrBackSide = Math.abs(angleDifference - Math.PI) < epsilon ? 1 : -1;
+            ModelRotationTuple rotatedModel = doorOrWindowRotatedModels.get(doorOrWindow);
+            if (rotatedModel != null 
+                && (missingModels.size() == 0 || !waitDoorOrWindowModelsLoadingEnd)) {
+              createGeometriesSurroundingDoorOrWindow(doorOrWindow, rotatedModelsFrontAreas.get(rotatedModel), frontOrBackSide,
+                  wall, wallGeometries, wallTopGeometries, 
+                  wallSidePoints, wallElevation, cosWallYawAngle, sinWallYawAngle, topLineAlpha, topLineBeta,
+                  texture, textureReferencePoint, wallSide);
+            } else {
+              missingModels.add(doorOrWindow);
+            }
+          }
+        }
+      }
+      if (missingModels.size() > 0) {
+        final ModelManager modelManager = ModelManager.getInstance();
+        for (final HomePieceOfFurniture doorOrWindow : (List<HomePieceOfFurniture>)missingModels.clone()) {
+          double angleDifference = Math.abs(wallYawAngle - doorOrWindow.getAngle()) % (2 * Math.PI);
+          final int frontOrBackSide = Math.abs(angleDifference - Math.PI) < epsilon ? 1 : -1;
+          // Load the model of the door or window to compute its front area  
+          modelManager.loadModel(doorOrWindow.getModel(), waitDoorOrWindowModelsLoadingEnd,
+              new ModelManager.ModelObserver() {
+                public void modelUpdated(BranchGroup modelRoot) {
+                  // Check again whether rotation model key and its front area weren't recently put in cache
+                  ModelRotationTuple rotatedModel = doorOrWindowRotatedModels.get(doorOrWindow);
+                  Area frontArea;
+                  if (rotatedModel == null) {
+                    rotatedModel = new ModelRotationTuple(doorOrWindow.getModel(), doorOrWindow.getModelRotation());
+                    frontArea = rotatedModelsFrontAreas.get(rotatedModel);
+                    if (frontArea == null) {
+                      // Add rotated piece model scene to a normalized transform group
+                      TransformGroup rotation = new TransformGroup(modelManager.getRotationTransformation(doorOrWindow.getModelRotation()));
+                      rotation.addChild(modelRoot);
+                      frontArea = modelManager.getFrontArea(((DoorOrWindow)doorOrWindow).getCutOutShape(), rotation);
+                      // Keep front area for a model with a given rotation in cache to avoid multiple calls to getFrontArea
+                      rotatedModelsFrontAreas.put(rotatedModel, frontArea);
+                    }
+                    // Keep rotated model key in cache for future updates
+                    doorOrWindowRotatedModels.put(doorOrWindow, rotatedModel);
+                  } else {
+                    frontArea = rotatedModelsFrontAreas.get(rotatedModel);
+                  }
+                  if (waitDoorOrWindowModelsLoadingEnd) {
+                    createGeometriesSurroundingDoorOrWindow(doorOrWindow, frontArea, frontOrBackSide,
+                        wall, wallGeometries, wallTopGeometries, 
+                        wallSidePoints, wallElevation, cosWallYawAngle, sinWallYawAngle, topLineAlpha, topLineBeta,
+                        texture, textureReferencePoint, wallSide);
+                  } else {
+                    missingModels.remove(doorOrWindow);
+                    if (missingModels.size() == 0) {
+                      EventQueue.invokeLater(new Runnable() {
+                          public void run() {
+                            // Request a new update only once all missing models are loaded
+                            updateWallSideGeometry(wallSide, texture, waitDoorOrWindowModelsLoadingEnd);
+                          }
+                        });
+                    }
+                  }
+                }
+                
+                public void modelError(Exception ex) {
+                  // In case of problem, ignore door or window geometry by using default full face cut out area
+                  ModelRotationTuple rotatedModel = new ModelRotationTuple(doorOrWindow.getModel(), doorOrWindow.getModelRotation());
+                  doorOrWindowRotatedModels.put(doorOrWindow, rotatedModel);
+                  if (rotatedModelsFrontAreas.get(rotatedModel) == null) {
+                    rotatedModelsFrontAreas.put(rotatedModel, FULL_FACE_CUT_OUT_AREA);
+                  }
+                  if (!waitDoorOrWindowModelsLoadingEnd) {
+                    missingModels.remove(doorOrWindow);
+                  }
+                }
+              });
+        }
+      }
+    }
   }
 
   /**
@@ -465,12 +582,12 @@ public class Wall3D extends Object3DBranch {
 
   /**
    * Returns the vertical rectangles that join each point of <code>points</code>
-   * and spread from <code>yMin</code> to a top line (y = ax + b) described by <code>topLineAlpha</code>
+   * and spread from <code>minElevation</code> to a top line (y = ax + b) described by <code>topLineAlpha</code>
    * and <code>topLineBeta</code> factors in a vertical plan that is rotated around
    * vertical axis matching <code>cosWallYawAngle</code> and <code>sinWallYawAngle</code>. 
    */
   private Geometry createWallVerticalPartGeometry(Wall wall, 
-                                                  float [][] points, float yMin, 
+                                                  float [][] points, float minElevation, 
                                                   double cosWallYawAngle, double sinWallYawAngle, 
                                                   double topLineAlpha, double topLineBeta, 
                                                   HomeTexture texture,
@@ -514,7 +631,7 @@ public class Wall3D extends Object3DBranch {
           textureReferencePoint [0] - arcCircleCenter [0]);
     }
     for (int i = 0; i < points.length; i++) {
-      bottom [i] = new Point3f(points [i][0], yMin, points [i][1]);
+      bottom [i] = new Point3f(points [i][0], minElevation, points [i][1]);
       if (arcCircleCenter == null) {
         distanceSqToWallMiddle [i] = Line2D.ptLineDistSq(xStart, yStart, xEnd, yEnd, bottom [i].x, bottom [i].z);
       } else {
@@ -523,8 +640,7 @@ public class Wall3D extends Object3DBranch {
         distanceSqToWallMiddle [i] *= distanceSqToWallMiddle [i];
       }
       // Compute vertical top point 
-      double xTopPointWithZeroYaw = cosWallYawAngle * points [i][0] + sinWallYawAngle * points [i][1];
-      float topY = (float)(topLineAlpha * xTopPointWithZeroYaw + topLineBeta);
+      float topY = getWallPointElevation(points [i][0], points [i][1], cosWallYawAngle, sinWallYawAngle, topLineAlpha, topLineBeta);
       top [i] = new Point3f(points [i][0], topY, points [i][1]);
     }
     // Search which rectangles should be ignored
@@ -549,7 +665,7 @@ public class Wall3D extends Object3DBranch {
     List<Point3f> coords = new ArrayList<Point3f> (rectanglesCount * 4);
     for (int index = 0; index < points.length; index++) {
       if (usedRectangle [index]) {
-        float y = yMin;
+        float y = minElevation;
         Point3f point1 = bottom [index];
         int nextIndex = (index + 1) % points.length;
         Point3f point2 = bottom [nextIndex];
@@ -577,8 +693,8 @@ public class Wall3D extends Object3DBranch {
     if (texture != null) {
       float halfThicknessSq = (wall.getThickness() * wall.getThickness()) / 4;
       TexCoord2f [] textureCoords = new TexCoord2f [coords.size()];
-      float yMinTextureCoords = yMin / texture.getHeight();
-      TexCoord2f firstTextureCoords = new TexCoord2f(0, yMinTextureCoords);
+      float minElevationTextureCoords = minElevation / texture.getHeight();
+      TexCoord2f firstTextureCoords = new TexCoord2f(0, minElevationTextureCoords);
       int j = 0;
       // Tolerate more error with round walls since arc points are approximative
       float epsilon = arcCircleCenter == null 
@@ -620,17 +736,17 @@ public class Wall3D extends Object3DBranch {
               secondHorizontalTextureCoords = -secondHorizontalTextureCoords;
             }
 
-            textureCoords1 = new TexCoord2f(firstHorizontalTextureCoords, yMinTextureCoords);
-            textureCoords2 = new TexCoord2f(secondHorizontalTextureCoords, yMinTextureCoords);
+            textureCoords1 = new TexCoord2f(firstHorizontalTextureCoords, minElevationTextureCoords);
+            textureCoords2 = new TexCoord2f(secondHorizontalTextureCoords, minElevationTextureCoords);
           } else {
             textureCoords1 = firstTextureCoords;
             float horizontalTextureCoords = (float)Point2D.distance(points [index][0], points [index][1], 
                 points [nextIndex][0], points [nextIndex][1]) / texture.getWidth();
-            textureCoords2 = new TexCoord2f(horizontalTextureCoords, yMinTextureCoords);
+            textureCoords2 = new TexCoord2f(horizontalTextureCoords, minElevationTextureCoords);
           }
           
           if (subpartSize > 0) {
-            float y = yMin;
+            float y = minElevation;
             for (float yMax = Math.min(top [index].y, top [nextIndex].y) - subpartSize / 2; y < yMax; y += subpartSize) {
               textureCoords [j++] = textureCoords1;
               textureCoords [j++] = textureCoords2;
@@ -660,6 +776,16 @@ public class Wall3D extends Object3DBranch {
     return geometryInfo.getIndexedGeometryArray();
   }
 
+  /**
+   * Returns the elevation of the wall at the given point.
+   */
+  private float getWallPointElevation(float xWallPoint, float yWallPoint, 
+                                      double cosWallYawAngle, double sinWallYawAngle,
+                                      double topLineAlpha, double topLineBeta) {
+    double xTopPointWithZeroYaw = cosWallYawAngle * xWallPoint + sinWallYawAngle * yWallPoint;
+    return (float)(topLineAlpha * xTopPointWithZeroYaw + topLineBeta);
+  }
+
   /**
    * Returns <code>pointAngle</code> plus or minus 2 PI to ensure <code>pointAngle</code> value 
    * will be greater or lower than <code>referencePointAngle</code> depending on <code>arcExtent</code> direction.
@@ -736,6 +862,220 @@ public class Wall3D extends Object3DBranch {
     return geometryInfo.getIndexedGeometryArray ();
   }
   
+  /**
+   * Creates the geometry surrounding the given non rectangular door or window. 
+   */
+  private void createGeometriesSurroundingDoorOrWindow(HomePieceOfFurniture doorOrWindow, 
+                                                       Area doorOrWindowFrontArea,
+                                                       float frontOrBackSide,
+                                                       Wall wall, List<Geometry> wallGeometries, 
+                                                       List<Geometry> wallTopGeometries,
+                                                       float [][] wallSidePoints, 
+                                                       float wallElevation, 
+                                                       double cosWallYawAngle, double sinWallYawAngle, 
+                                                       double topLineAlpha, double topLineBeta,
+                                                       HomeTexture texture, float [] textureReferencePoint, 
+                                                       int wallSide) {
+    Area fullFaceArea = new Area(FULL_FACE_CUT_OUT_AREA);
+    fullFaceArea.subtract(doorOrWindowFrontArea);
+    if (!fullFaceArea.isEmpty()) {
+      // Compute the depth position of the surrounding wall part in normalized piece space  
+      float doorOrWindowDepth = doorOrWindow.getDepth();
+      float xPieceSide = (float)(doorOrWindow.getX() - frontOrBackSide * doorOrWindowDepth / 2 * Math.sin(doorOrWindow.getAngle()));
+      float yPieceSide = (float)(doorOrWindow.getY() + frontOrBackSide * doorOrWindowDepth / 2 * Math.cos(doorOrWindow.getAngle()));
+      float [] wallFirstPoint = wallSide == WALL_LEFT_SIDE 
+          ? wallSidePoints [0]
+          : wallSidePoints [wallSidePoints.length - 1];
+      float [] wallSecondPoint = wallSide == WALL_LEFT_SIDE 
+          ? wallSidePoints [wallSidePoints.length / 2 - 1]
+          : wallSidePoints [wallSidePoints.length / 2];
+      float frontSideToWallDistance = (float)Line2D.ptLineDist(wallFirstPoint [0], wallFirstPoint [1], 
+              wallSecondPoint [0], wallSecondPoint [1], xPieceSide, yPieceSide);
+      float position = (float)Line2D.relativeCCW(wallFirstPoint [0], wallFirstPoint [1], 
+          wallSecondPoint [0], wallSecondPoint [1], xPieceSide, yPieceSide);
+      float depthTranslation = frontOrBackSide * (0.5f - position * frontSideToWallDistance / doorOrWindowDepth);
+      
+      // Compute surrounding part transformation matrix
+      Transform3D frontAreaTransform = ModelManager.getInstance().getPieceOFFurnitureNormalizedModelTransformation(doorOrWindow);       
+      Transform3D frontAreaTranslation = new Transform3D();
+      frontAreaTranslation.setTranslation(new Vector3f(0, 0, depthTranslation));
+      frontAreaTransform.mul(frontAreaTranslation);
+  
+      // Generate wall side path in normalized piece space
+      Transform3D invertedFrontAreaTransform = new Transform3D();
+      invertedFrontAreaTransform.invert(frontAreaTransform);    
+      GeneralPath wallPath = new GeneralPath();
+      Point3f wallPoint = new Point3f(wallFirstPoint [0], wallElevation, wallFirstPoint [1]);
+      invertedFrontAreaTransform.transform(wallPoint);
+      wallPath.moveTo(wallPoint.x, wallPoint.y);
+      wallPoint = new Point3f(wallSecondPoint [0], wallElevation, wallSecondPoint [1]);
+      invertedFrontAreaTransform.transform(wallPoint);
+      wallPath.lineTo(wallPoint.x, wallPoint.y);
+      Point3f topWallPoint1 = new Point3f(wallSecondPoint [0], 
+          getWallPointElevation(wallSecondPoint [0], wallSecondPoint [1], 
+              cosWallYawAngle, sinWallYawAngle, topLineAlpha, topLineBeta), wallSecondPoint [1]);
+      invertedFrontAreaTransform.transform(topWallPoint1);
+      wallPath.lineTo(topWallPoint1.x, topWallPoint1.y);
+      Point3f topWallPoint2 = new Point3f(wallFirstPoint [0], 
+          getWallPointElevation(wallFirstPoint [0], wallFirstPoint [1], 
+              cosWallYawAngle, sinWallYawAngle, topLineAlpha, topLineBeta), wallFirstPoint [1]);
+      invertedFrontAreaTransform.transform(topWallPoint2);
+      wallPath.lineTo(topWallPoint2.x, topWallPoint2.y);
+      wallPath.closePath();
+  
+      GeneralPath doorOrWindowSurroundingPath = new GeneralPath();
+      doorOrWindowSurroundingPath.moveTo(-.5f, -.5f);
+      doorOrWindowSurroundingPath.lineTo(-.5f,  .5f);
+      doorOrWindowSurroundingPath.lineTo( .5f, .5f);
+      doorOrWindowSurroundingPath.lineTo( .5f, -.5f);
+      doorOrWindowSurroundingPath.closePath();
+      
+      // Retrieve the points of the surrounding area as polygons where holes are removed
+      Area doorOrWindowSurroundingArea = new Area(doorOrWindowSurroundingPath);
+      doorOrWindowSurroundingArea.intersect(new Area(wallPath));
+      doorOrWindowSurroundingArea.subtract(doorOrWindowFrontArea);
+      float flatness = 0.5f / (Math.max(doorOrWindow.getWidth(), doorOrWindow.getHeight()));
+      if (!doorOrWindowSurroundingArea.isEmpty()) {
+        boolean reversed = frontOrBackSide > 0 ^ wallSide == WALL_RIGHT_SIDE ^ doorOrWindow.isModelMirrored();
+        List<float [][]> doorOrWindowSurroundingAreasPoints = getAreaPoints(doorOrWindowSurroundingArea, flatness, reversed);     
+        if (!doorOrWindowSurroundingAreasPoints.isEmpty()) {
+          // Generate coordinates for the surrounding area, door or window border and missing top wall part in the wall thickness
+          int [] stripCounts = new int [doorOrWindowSurroundingAreasPoints.size()];
+          int vertexCount = 0;
+          for (int i = 0; i < doorOrWindowSurroundingAreasPoints.size(); i++) {
+            float [][] areaPoints = doorOrWindowSurroundingAreasPoints.get(i);
+            stripCounts [i] = areaPoints.length + 1;
+            vertexCount += stripCounts [i]; 
+          }
+          float halfWallThickness = wall.getThickness() / 2;
+          float deltaXToWallMiddle = (float)(halfWallThickness * sinWallYawAngle);
+          float deltaZToWallMiddle = -(float)(halfWallThickness * cosWallYawAngle);
+          if (wallSide == WALL_LEFT_SIDE) {
+            deltaXToWallMiddle *= -1;
+            deltaZToWallMiddle *= -1;
+          }
+          Point3f [] coords = new Point3f [vertexCount];
+          List<Point3f> borderCoords = new ArrayList<Point3f>(4 * vertexCount);
+          List<Point3f> slopingTopCoords = new ArrayList<Point3f>();        
+          TexCoord2f [] textureCoords;
+          List<TexCoord2f> borderTextureCoords;
+          if (texture != null) {
+            textureCoords = new TexCoord2f [coords.length];
+            borderTextureCoords = new ArrayList<TexCoord2f>(4 * vertexCount);
+          } else {
+            textureCoords = null;
+            borderTextureCoords = null;
+          }
+          int i = 0;
+          for (float [][] areaPoints : doorOrWindowSurroundingAreasPoints) {
+            Point3f point = new Point3f(areaPoints [0][0], areaPoints [0][1], 0);
+            frontAreaTransform.transform(point);
+            TexCoord2f textureCoord = null;
+            if (texture != null) {
+              // Compute texture coordinates of wall side according to textureReferencePoint
+              float horizontalTextureCoords = (float)Point2D.distance(textureReferencePoint[0], textureReferencePoint[1], 
+                  point.x, point.z) / texture.getWidth();
+              if (wallSide == WALL_LEFT_SIDE && texture.isLeftToRightOriented()) {
+                horizontalTextureCoords = -horizontalTextureCoords;
+              }
+              textureCoord = new TexCoord2f(horizontalTextureCoords, point.y / texture.getHeight());
+            }
+            double distanceToTop = Line2D.ptLineDistSq(topWallPoint1.x, topWallPoint1.y, topWallPoint2.x, topWallPoint2.y, 
+                areaPoints [0][0], areaPoints [0][1]);
+            
+            for (int j = 0; j < areaPoints.length; j++, i++) {
+              // Store coordinates of the surrounding area
+              coords [i] = point;
+              if (texture != null) {
+                textureCoords [i] = textureCoord;
+              }
+              
+              // Generate coordinates of the door or window border or missing top wall part
+              int nextPointIndex = j < areaPoints.length - 1  
+                  ? j + 1
+                  : 0;
+              // Select the coordinates list to which the next quadrilateral will be added
+              List<Point3f> coordsList;
+              double nextDistanceToTop = Line2D.ptLineDistSq(topWallPoint1.x, topWallPoint1.y, topWallPoint2.x, topWallPoint2.y, 
+                  areaPoints [nextPointIndex][0], areaPoints [nextPointIndex][1]);
+              if (distanceToTop < 1E-10 && nextDistanceToTop < 1E-10) {
+                coordsList = slopingTopCoords;
+              } else {
+                coordsList = borderCoords;
+              }
+              
+              Point3f nextPoint = new Point3f(areaPoints [nextPointIndex][0], areaPoints [nextPointIndex][1], 0);
+              frontAreaTransform.transform(nextPoint);            
+              coordsList.add(point);
+              coordsList.add(new Point3f(point.x + deltaXToWallMiddle, point.y, point.z + deltaZToWallMiddle));
+              coordsList.add(new Point3f(nextPoint.x + deltaXToWallMiddle, nextPoint.y, nextPoint.z + deltaZToWallMiddle));
+              coordsList.add(nextPoint);
+              
+              TexCoord2f nextTextureCoord = null;
+              if (texture != null) {
+                float horizontalTextureCoords = (float)Point2D.distance(textureReferencePoint[0], textureReferencePoint[1], 
+                    nextPoint.x, nextPoint.z) / texture.getWidth();
+                if (wallSide == WALL_LEFT_SIDE && texture.isLeftToRightOriented()) {
+                  horizontalTextureCoords = -horizontalTextureCoords;
+                }
+                nextTextureCoord = new TexCoord2f(horizontalTextureCoords, nextPoint.y / texture.getHeight());
+                if (coordsList == borderCoords) {
+                  borderTextureCoords.add(textureCoord);
+                  borderTextureCoords.add(textureCoord);
+                  borderTextureCoords.add(nextTextureCoord);
+                  borderTextureCoords.add(nextTextureCoord);
+                }
+              }
+              
+              // Update values for next loop round
+              distanceToTop = nextDistanceToTop;
+              point = nextPoint;
+              textureCoord = nextTextureCoord;
+            }
+            
+            // Close the polygon with first point for special cases
+            coords [i] = point;
+            if (texture != null) {
+              textureCoords [i] = textureCoord;
+            }
+            i++;
+          }
+          
+          // Generate surrounding area geometry
+          GeometryInfo geometryInfo = new GeometryInfo(GeometryInfo.POLYGON_ARRAY);
+          geometryInfo.setStripCounts(stripCounts);
+          geometryInfo.setCoordinates(coords);
+          if (texture != null) {
+            geometryInfo.setTextureCoordinateParams(1, 2);
+            geometryInfo.setTextureCoordinates(0, textureCoords);
+          }
+          new NormalGenerator().generateNormals(geometryInfo);
+          wallGeometries.add(geometryInfo.getIndexedGeometryArray());
+        
+          if (borderCoords.size() > 0) { 
+            // Generate border geometry 
+            geometryInfo = new GeometryInfo(GeometryInfo.QUAD_ARRAY);        
+            geometryInfo.setCoordinates(borderCoords.toArray(new Point3f [borderCoords.size()]));          
+            if (texture != null) {
+              geometryInfo.setTextureCoordinateParams(1, 2);
+              geometryInfo.setTextureCoordinates(0, borderTextureCoords.toArray(new TexCoord2f [borderTextureCoords.size()]));
+            }
+            new NormalGenerator(Math.PI / 2).generateNormals(geometryInfo);
+            wallGeometries.add(geometryInfo.getIndexedGeometryArray());
+          }
+          
+          if (slopingTopCoords.size() > 0) { 
+            // Generate wall top geometry 
+            geometryInfo = new GeometryInfo(GeometryInfo.QUAD_ARRAY);        
+            geometryInfo.setCoordinates(slopingTopCoords.toArray(new Point3f [slopingTopCoords.size()]));
+            new NormalGenerator().generateNormals(geometryInfo);
+            wallTopGeometries.add(geometryInfo.getIndexedGeometryArray());
+          }
+        }
+      }
+    }
+  }
+
   /**
    * Returns the elevation of the wall managed by this 3D object.
    */
@@ -914,4 +1254,46 @@ public class Wall3D extends Object3DBranch {
       return this.doorsOrWindows;
     }
   }
+
+  /**
+   * A class used to store model and its rotation as a key.
+   * @author Emmanuel Puybaret
+   */
+  private static class ModelRotationTuple {
+    private Content    model;
+    private float [][] rotation;
+
+    public ModelRotationTuple(Content model, float [][] rotation) {
+      this.model = model;
+      this.rotation = rotation;
+    }
+    
+    @Override
+    public int hashCode() {
+      int hashCode = 31 * this.model.hashCode();
+      for (float [] table : this.rotation) {
+        hashCode += Arrays.hashCode(table);
+      }
+      return hashCode;
+    }
+    
+    @Override
+    public boolean equals(Object obj) {
+      if (this == obj) {
+        return true;
+      } else if (obj instanceof ModelRotationTuple) {
+        ModelRotationTuple tuple = (ModelRotationTuple)obj;
+        if (this.model.equals(tuple.model)
+            && this.rotation.length == tuple.rotation.length) {
+          for (int i = 0; i < this.rotation.length; i++) {
+            if (!Arrays.equals(this.rotation [i], tuple.rotation [i])) {
+              return false;
+            }
+          }
+          return true;
+        }
+      }
+      return false;
+    }
+  }
 }
\ No newline at end of file
