@@ -50,7 +50,6 @@ import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collections;
 import java.util.List;
-import java.util.regex.Pattern;
 
 import javax.swing.AbstractAction;
 import javax.swing.AbstractListModel;
@@ -343,6 +342,12 @@ class FurnitureCatalogListPanel extends JPanel implements View {
                                  final FurnitureCatalogController controller) {
     final Cursor handCursor = new Cursor(Cursor.HAND_CURSOR);
     MouseInputAdapter mouseListener = new MouseInputAdapter () {
+        // Track first drag and drop under Mac OS X / Java 7 where the selection in the catalog
+        // visually changes the first time the cursor enters in the plan component   
+        private boolean                 firstMouseDrag = OperatingSystem.isMacOSXLionOrSuperior()
+                                                         && OperatingSystem.isJavaVersionGreaterOrEqual("1.7");;
+        private CatalogPieceOfFurniture firstSelectedPiece;
+        
         @Override
         public void mouseClicked(MouseEvent ev) {
           if (SwingUtilities.isLeftMouseButton(ev)) {
@@ -374,6 +379,35 @@ class FurnitureCatalogListPanel extends JPanel implements View {
           }
           setCursor(Cursor.getDefaultCursor());
         }
+        
+        @Override
+        public void mouseDragged(MouseEvent ev) {
+          if (this.firstMouseDrag
+              && this.firstSelectedPiece != null) {
+            List<CatalogPieceOfFurniture> selectedFurniture = controller.getSelectedFurniture();
+            if (!selectedFurniture.isEmpty()
+                && this.firstSelectedPiece != selectedFurniture.get(0)) {
+              controller.setSelectedFurniture(Arrays.asList(this.firstSelectedPiece));
+              this.firstMouseDrag = false;
+              this.firstSelectedPiece = null;
+            }
+          }
+        }
+        
+        @Override
+        public void mousePressed(MouseEvent ev) {
+          if (this.firstMouseDrag) {
+            EventQueue.invokeLater(new Runnable() {
+                public void run() {
+                  List<CatalogPieceOfFurniture> selectedFurniture = controller.getSelectedFurniture();
+                  if (!selectedFurniture.isEmpty()) {
+                    // Store the selected piece at the first drag and drop
+                    firstSelectedPiece = selectedFurniture.get(0);
+                  }
+                }
+              });
+          }
+        }
       };
     catalogFurnitureList.addMouseListener(mouseListener);
     catalogFurnitureList.addMouseMotionListener(mouseListener);
@@ -744,15 +778,15 @@ class FurnitureCatalogListPanel extends JPanel implements View {
    * List model adaptor to CatalogPieceOfFurniture instances of catalog.  
    */
   private static class CatalogListModel extends AbstractListModel {
-    private FurnitureCatalog              catalog;
-    private List<CatalogPieceOfFurniture> furniture;
-    private FurnitureCategory             filterCategory;
-    private Pattern                       filterNamePattern;
+    private FurnitureCatalog                catalog;
+    private List<CatalogPieceOfFurniture>   furniture;
+    private FurnitureCategory               filterCategory;
+    private String                          filterText;
     
     public CatalogListModel(FurnitureCatalog catalog) {
       this.catalog = catalog;
       this.furniture = new ArrayList<CatalogPieceOfFurniture>();
-      this.filterNamePattern = Pattern.compile(".*");
+      this.filterText = "";
       catalog.addFurnitureListener(new CatalogFurnitureListener(this));
       updateFurnitureList();
     }
@@ -763,7 +797,7 @@ class FurnitureCatalogListPanel extends JPanel implements View {
     }
 
     public void setFilterText(String filterText) {
-      this.filterNamePattern = Pattern.compile(".*" + filterText + ".*", Pattern.CASE_INSENSITIVE);
+      this.filterText = filterText;
       updateFurnitureList();
     }
 
@@ -779,21 +813,10 @@ class FurnitureCatalogListPanel extends JPanel implements View {
       this.furniture.clear();
       for (FurnitureCategory category : this.catalog.getCategories()) {
         for (CatalogPieceOfFurniture piece : category.getFurniture()) {
-          if (this.filterCategory == null
-               || piece.getCategory().equals(this.filterCategory)) {
-            if (this.filterNamePattern.matcher(piece.getName()).matches()
-                 || this.filterNamePattern.matcher(piece.getCategory().getName()).matches()                   
-                 || (piece.getCreator() != null && this.filterNamePattern.matcher(piece.getCreator()).matches())
-                 || (piece.getDescription() != null && this.filterNamePattern.matcher(piece.getDescription()).matches())) {
-              this.furniture.add(piece);
-            } else {
-              for (String tag : piece.getTags()) {
-                if (this.filterNamePattern.matcher(tag).matches()) {
-                  this.furniture.add(piece);
-                  break;
-                }
-              }
-            }
+          if ((this.filterCategory == null
+                || piece.getCategory().equals(this.filterCategory))
+              && piece.matchesFilter(this.filterText)) {
+            furniture.add(piece);
           }
         }
       }
