@@ -269,26 +269,33 @@ public class ImportedFurnitureWizardStepsPanel extends JPanel
         
         @Override
         public boolean importData(JComponent comp, Transferable transferedFiles) {
-          boolean success = true;
+          boolean success = false;
           try {
             List<File> files = (List<File>)transferedFiles.getTransferData(DataFlavor.javaFileListFlavor);
-            final String modelName = files.get(0).getAbsolutePath();
-            EventQueue.invokeLater(new Runnable() {
-                public void run() {
-                  updateController(modelName, preferences, 
-                      controller.getContentManager(), defaultModelCategory, false);
-                }
-              });
+            for (File file : files) {
+              final String modelName = file.getAbsolutePath();
+              // Try to import the first file that would be accepted by content manager
+              if (controller.getContentManager().isAcceptable(modelName, ContentManager.ContentType.MODEL)) {
+                EventQueue.invokeLater(new Runnable() {
+                    public void run() {
+                      updateController(modelName, preferences, 
+                          controller.getContentManager(), defaultModelCategory, false);
+                    }
+                  });
+                success = true;
+                break;
+              }
+            }
           } catch (UnsupportedFlavorException ex) {
-            success = false;
+            // No success
           } catch (IOException ex) {
-            success = false;
+            // No success
           }
           if (!success) {
             EventQueue.invokeLater(new Runnable() {
                 public void run() {
                   JOptionPane.showMessageDialog(SwingUtilities.getRootPane(ImportedFurnitureWizardStepsPanel.this), 
-                      preferences.getLocalizedString(ImportedFurnitureWizardStepsPanel.class, "modelChoiceError"));
+                      preferences.getLocalizedString(ImportedFurnitureWizardStepsPanel.class, "modelChoiceErrorLabel.text"));
                 }
               });
           }
@@ -463,17 +470,20 @@ public class ImportedFurnitureWizardStepsPanel extends JPanel
           name = name.trim();
           // If category is empty, replace it by the last selected item
           if (name.length() == 0) {
-            setItem(categoryComboBox.getSelectedItem());
-          } 
-          FurnitureCategory category = new FurnitureCategory(name);
-          // Search an existing category
-          List<FurnitureCategory> categories = preferences.getFurnitureCatalog().getCategories();
-          int categoryIndex = Collections.binarySearch(categories, category);
-          if (categoryIndex >= 0) {
-            return categories.get(categoryIndex);
+            Object selectedItem = categoryComboBox.getSelectedItem();
+            setItem(selectedItem);
+            return selectedItem;
+          } else {
+            FurnitureCategory category = new FurnitureCategory(name);
+            // Search an existing category
+            List<FurnitureCategory> categories = preferences.getFurnitureCatalog().getCategories();
+            int categoryIndex = Collections.binarySearch(categories, category);
+            if (categoryIndex >= 0) {
+              return categories.get(categoryIndex);
+            }
+            // If no existing category was found, return a new one          
+            return category;
           }
-          // If no existing category was found, return a new one          
-          return category;
         }
       
         public void setItem(Object value) {
@@ -1154,7 +1164,7 @@ public class ImportedFurnitureWizardStepsPanel extends JPanel
           try {
             // Copy model content to a temporary content
             modelContent = TemporaryURLContent.copyToTemporaryURLContent(modelContent);
-          } catch (IOException ex2) {
+          } catch (IOException ex) {
             setDefaultStateAndShowModelChoiceError(modelName, preferences, !ignoreException);
             return;
           }
@@ -1227,12 +1237,16 @@ public class ImportedFurnitureWizardStepsPanel extends JPanel
           }
           
           // Found no readable model
-          if (isShowing()) {
-            setDefaultState();
-            setModelChoiceTexts(preferences);
-            JOptionPane.showMessageDialog(SwingUtilities.getRootPane(ImportedFurnitureWizardStepsPanel.this), 
-                preferences.getLocalizedString(ImportedFurnitureWizardStepsPanel.class, "modelChoiceFormatError"));
-          }
+          EventQueue.invokeLater(new Runnable() {
+              public void run() {
+                if (isShowing()) {
+                  setDefaultState();
+                  setModelChoiceTexts(preferences);
+                  JOptionPane.showMessageDialog(SwingUtilities.getRootPane(ImportedFurnitureWizardStepsPanel.this), 
+                      preferences.getLocalizedString(ImportedFurnitureWizardStepsPanel.class, "modelChoiceFormatError"));
+                }
+              }
+            });
         }
       });
   }
